/*
	Copyright 2009 Michael Daffin <james1479@googlemail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QTextStream>

#include "dice.h"
#include "dicelist.h"
#include "dicepicker.h"
#include "mainwindow.h"
#include <cmath>

/* Qt */
#include <QWidget>
#include <QHBoxLayout>
#include <QLabel>
#include <QHBoxLayout>
#include <QList>
#include <QVector>
#include <QPushButton>
#include <QRegExpValidator>
#include <QCheckBox>
#include <QStringList>
#include <QComboBox>
#include <QDataStream>

/* kde */
#include <KIntSpinBox>
#include <KLineEdit>
#include <KIcon>
#include <KLocale>
/*
 *@return returns a list of the results rolled. The first is the total and the last is the modifier (c)
 */
QList<int> rollDice(int a, int b, int c)
{
	QList <int> list;
	for(int i = 0;i<a;++i)
			list.append(qrand() % b + 1);
	int total = 0;
	foreach(int i, list)
			total += i;
	total += c;
	list.prepend(total);
	list.append(c);
	return list;
}

/*
 *@return a string containing "value", with the bold html tags around it if value is greater or equal to "min".
 */
QString boldOn(int value, int min = 20)
{
	QString text; text = QString("%1").arg(value);

	if ( min <= 0 )
		return text;

	if ( value >= min )
	{
		text.prepend("<b>");
		text.append("</b>");
	}

	return text;
}

/*
 *@return returns a string in the format "a = (b + c + ... ) + d" with a being the first value in diceRolls, b the second...
 */
QString formatLine(QList<int> diceRolls, int max = 20)
{
	QString line;
	int mod = diceRolls.takeLast();
	line.append(QString("%1").arg(diceRolls.takeFirst()));
	line.append(" = (");
	for (int r = 0; r < diceRolls.count() - 1; ++r)
	{
		line.append(boldOn(diceRolls[r], max));
		line.append(" + ");
	}
	line.append(boldOn(diceRolls[diceRolls.count()-1], max));
	line.append(")");
	if ( mod > 0 )
		line.append(QString(" + %1").arg(mod));
	else if ( mod < 0 )
		line.append(QString(" - %1").arg(-mod));
	return line;
}

AbstractDice::AbstractDice(QWidget *parent)
					: QFrame(parent)
{
	deleteButton = new QPushButton(KIcon("list-remove"), "");
	rollButton = new QPushButton(tr("Roll"));

	rollButton->setWhatsThis(tr("Rolls the dice on the same row (or frame)"));
	deleteButton->setWhatsThis(tr("Deletes the dice on the same row (or frame)"));

	connect(rollButton, SIGNAL(clicked(bool)), this, SLOT(roll()));
	connect(deleteButton, SIGNAL(clicked(bool)), this, SIGNAL(deleteMe()));
}

CustemDice::CustemDice(QWidget *parent) : AbstractDice(parent)
{
	numberOfDice	= new KIntSpinBox(1, 999, 1, 1, this);
	numberOfSides	= new KIntSpinBox(1, 999, 1, 20, this);
	modifier		= new KIntSpinBox(-999, 999, 1, 0, this);
	dLabel			= new QLabel(tr("d"));
	modLabel		= new QLabel(tr("+"));

	numberOfDice->setWhatsThis(tr("The number of dice to roll"));
	numberOfSides->setWhatsThis(tr("The number of sides the dice has<br><br>"
								   "d20 is equlivant to rolling a dice with 20 sides"
								   " producing a number between 1 and 20"));
	modifier->setWhatsThis(tr("The ammount to add to the result of the dice roll"));

	mainLayout = new QHBoxLayout;

	mainLayout->addWidget(numberOfDice, 1);
	mainLayout->addWidget(dLabel, 0);
	mainLayout->addWidget(numberOfSides, 1);
	mainLayout->addWidget(modLabel, 0);
	mainLayout->addWidget(modifier, 1);
	mainLayout->addWidget(rollButton, 0);
	mainLayout->addWidget(deleteButton, 0);

	setLayout(mainLayout);
}

void CustemDice::roll()
{
	int dice	= numberOfDice->value();
	int sides	= numberOfSides->value();
	int mod		= modifier->value();
	QList<int> list(rollDice(dice, sides, mod));
	QString result = formatLine(list, sides);

	emit diceRolled(result);
}

void CustemDice::save(QDataStream& stream)
{
	stream << qint32(Custem);
	stream << qint32(numberOfDice->value());
	stream << qint32(numberOfSides->value());
	stream << qint32(modifier->value());
}

void CustemDice::open(QDataStream& stream)
{
	qint32 value;
	stream >> value;
	numberOfDice->setValue(value);
	stream >> value;
	numberOfSides->setValue(value);
	stream >> value;
	modifier->setValue(value);
}

D20Dice::D20Dice(QWidget *parent) : AbstractDice(parent)
{
	label		= new QLabel(tr("1d20+"));
	modifier	= new KIntSpinBox(-999, 999, 1, 0, this);

	label->setWhatsThis(tr("Represents a 20 sided dice"));
	modifier->setWhatsThis(tr("The ammount to add to the result of the dice roll"));

	mainLayout = new QHBoxLayout;

	mainLayout->addWidget(label, 0);
	mainLayout->addWidget(modifier, 1);
	mainLayout->addWidget(rollButton, 0);
	mainLayout->addWidget(deleteButton, 0);

	setLayout(mainLayout);
}


void D20Dice::roll()
{
	int dice	= 1;
	int sides	= 20;
	int mod		= modifier->value();

	QList<int> list(rollDice(dice, sides, mod));
	QString result = formatLine(list, sides);

	emit diceRolled(result);
}

void D20Dice::save(QDataStream& stream)
{
	stream << qint32(D20);
	stream << qint32(modifier->value());
}

void D20Dice::open(QDataStream& stream)
{
	qint32 value;
	stream >> value;
	modifier->setValue(value);
}

AttackDamageDice::AttackDamageDice(QWidget *parent) : AbstractDice(parent)
{
	numberOfDice	= new KIntSpinBox(1, 999, 1, 1, this);
	dLabel			= new QLabel("d");
	diceSize		= new KIntSpinBox(1, 999, 1, 8, this);
	modLabel		= new QLabel("+");
	modifier		= new KIntSpinBox(-999, 999, 1, 0, this);
	critical		= new QCheckBox;

	numberOfDice->setWhatsThis(tr("The number of dice to roll"));
	diceSize->setWhatsThis(tr("The number of sides the dice has<br><br>"
								   "d20 is equlivant to rolling a dice with 20 sides"
								   " producing a number between 1 and 20"));
	modifier->setWhatsThis(tr("The ammount to add to the result of the dice roll"));
	critical->setWhatsThis(tr("Sets whether or not this dice set is affected by a critical hit."
							  "<br><br>If set to true the result of this dice will be multiplyed on"
							  " a critical hit."
							  "<br><br>If set to false the result will never be multiplied on "
							  "a ctirical hit"));

	critical->setCheckState(Qt::Checked);

	mainLayout = new QHBoxLayout;
	setMargin(0);

	mainLayout->addWidget(numberOfDice, 1);
	mainLayout->addWidget(dLabel, 0);
	mainLayout->addWidget(diceSize, 1);
	mainLayout->addWidget(modLabel, 0);
	mainLayout->addWidget(modifier, 1);
	mainLayout->addWidget(critical, 0);
	mainLayout->addWidget(deleteButton, 0);

	setLayout(mainLayout);

	connect(deleteButton, SIGNAL(clicked()), this, SIGNAL(deleteMe()));
}

QList<int> AttackDamageDice::rollDamage()
{
	int dice	= numberOfDice->value();
	int sides	= diceSize->value();
	int mod		= modifier->value();
	QList<int> list(rollDice(dice, sides, mod));

	return list;
}

void AttackDamageDice::save(QDataStream& stream)
{
	stream << qint32(numberOfDice->value());
	stream << qint32(diceSize->value());
	stream << qint32(modifier->value());
	stream << critical->isChecked();
}

void AttackDamageDice::open(QDataStream& stream)
{
	qint32 value;
	stream >> value;
	numberOfDice->setValue(value);
	stream >> value;
	diceSize->setValue(value);
	stream >> value;
	modifier->setValue(value);

	bool checked;
	stream >> checked;
	critical->setChecked(checked);
}

AttackDice::AttackDice(QWidget *parent) : AbstractDice(parent)
{
	attackBonusLabel	= new QLabel("Attack Bonus: ");
	attackBonus			= new KLineEdit("1");
	damageLabel			= new QLabel("Damage: ");
	critLabel			= new QLabel("Crit?");
	addDamageDiceButton	= new QPushButton(KIcon("list-add"), "");
	critRangeLabel		= new QLabel("Crit Range: ");
	critRange			= new KIntSpinBox(1, 20, 1, 20, this);
	multiplier			= new KIntSpinBox(1, 10, 1, 2, this);

	attackBonus->setWhatsThis(tr("Your attack bonus in the format attack1/attack2/attack3..."
								 "<br><br>This will be used to determin the number of attacks to roll "
								 "and the ammount to modify each attack by."));
	addDamageDiceButton->setWhatsThis(tr("Adds an extra dice for damage"));
	critRange->setWhatsThis(tr("This is the critical range, you should enter the lower "
							   "value eg. if your critical range is 18-20 enter 18"));
	multiplier->setWhatsThis(tr("The damage multiplier for critical hits.<br><br>"
								"The result of each dice will be multiplied by this value"
								"on a critical hit if that dice can be criticaled"));

	QRegExp rx("(\\d/?){1,99}");
	QRegExpValidator *attackBonusValidator = new QRegExpValidator(rx, this);
	attackBonus->setValidator(attackBonusValidator);
	attackBonus->setAlignment(Qt::AlignCenter);
	multiplier->setPrefix("X");

	mainLayout = new QVBoxLayout;

	QHBoxLayout *line = new QHBoxLayout;
	line->addWidget(attackBonusLabel, 0);
	line->addWidget(attackBonus, 1);
	line->addWidget(deleteButton, 0);
	mainLayout->addLayout(line);

	line = new QHBoxLayout;
	line->addWidget(damageLabel, 1);
	line->addWidget(critLabel, 0);
	line->addWidget(addDamageDiceButton, 0);
	mainLayout->addLayout(line);

	line = new QHBoxLayout;
	line->addWidget(critRangeLabel, 0);
	line->addWidget(critRange, 1);
	line->addWidget(multiplier, 1);
	line->addWidget(rollButton, 0);
	mainLayout->addLayout(line);

	addDamageDice();

	setLayout(mainLayout);
	setFrameStyle(QFrame::Raised | QFrame::StyledPanel);

	connect(addDamageDiceButton, SIGNAL(clicked()), this, SLOT(addDamageDice()));
}


void AttackDice::roll()
{
	QString result;
	QString ab(attackBonus->text());
	if (ab.isEmpty())
		ab.append("0");
	QStringList attackList(ab.split("/"));
	result.append("--------------------");
	for (int i = 0; i < attackList.count(); ++i)
	{
		QString attack(attackList[i]);
		QList<int> toHit = rollDice(1, 20, attack.toInt());
		bool critical = false;

		result.append(QString("<br>Attack %1: ").arg(i+1));
		result.append(formatLine(toHit, critRange->value()));
		if (toHit[1] >= critRange->value())
		{
			result.append(tr(" <b>Critical!</b><br>"));
			result.append(tr("Confrim: "));
			QList<int> toHit = rollDice(1, 20, attack.toInt());
			result.append(formatLine(toHit, 20));
			critical = true;
		}
		if (toHit[1] <= 1)
		{
			result.append(tr(" <b>Natural 1!</b><br>"));
			result.append(tr("Confrim: "));
			QList<int> toHit = rollDice(1, 20, 0);
			result.append(formatLine(toHit, 20));
			result.append(tr("<br>"));
			continue;
		}
		result.append("<br>");
		QList<int> rolledDamage[damageDiceList.count()];

		QList<int> criticalList;

		for(int d = 0; d < damageDiceList.count(); ++d)
		{
			rolledDamage[d] = damageDiceList[d]->rollDamage();
			result.append(tr("Damage: "));
			if (critical == true) {
				if (damageDiceList[d]->isCritable())
				{
					criticalList.append(rolledDamage[d][0] * multiplier->value());
					result.append(QString("%1 = (%2 X) ").arg(criticalList[d])
								  .arg(multiplier->value()));
				}
				else
					criticalList.append(rolledDamage[d][0]);
			}

			result.append(formatLine(rolledDamage[d], 0));
			result.append("<br>");
		}
		int total = 0;
		result.append(tr("Total Damage: "));
		if (critical)
		{
			for (int j = 0; j < damageDiceList.count(); ++j)
				total += criticalList[j];
			result.append(tr(" Critical: %1, ").arg(total));
		}
		total = 0;
		for (int j = 0; j < damageDiceList.count(); ++j)
			total += rolledDamage[j][0];
		if (critical)
			result.append(tr("Normal: "));
		result.append(QString("%1").arg(total));
		result.append("<br>");
	}
	result.append("--------------------");

	emit diceRolled(result);
}

void AttackDice::save(QDataStream& stream)
{
	stream << qint32(Attack);

	stream << attackBonus->text();
	stream << qint32(critRange->value());
	stream << qint32(multiplier->value());

	stream << qint32(damageDiceList.count());
	foreach(AttackDamageDice* dice, damageDiceList)
		dice->save(stream);
}

void AttackDice::open(QDataStream& stream)
{
	qint32 value;
	QString text;

	stream >> text;
	attackBonus->setText(text);
	stream >> value;
	critRange->setValue(value);
	stream >> value;
	multiplier->setValue(value);

	stream >> value;
	int difference = 0;
	difference = fabs(damageDiceList.count() - value);

	if(damageDiceList.count() < value )
	{
		for(int i = 0; i < difference; ++i)
			addDamageDice();
	}
	else if(damageDiceList.count() > value)
	{
		for(int i = 0; i < difference; ++i)
			removeDamageDice();
	}

	foreach(AttackDamageDice* dice, damageDiceList)
		dice->open(stream);
}

void AttackDice::addDamageDice()
{
	int length = damageDiceList.count();

	if (length == 5)
		return;

	AttackDamageDice* dice = new AttackDamageDice(this);
	damageDiceList.append( dice );

	if (length == 4)
		addDamageDiceButton->setEnabled(false);
	if ( length == 0 )
		damageDiceList[length]->setDeleteEnabled(false);
	else
		damageDiceList[0]->setDeleteEnabled(true);
	mainLayout->insertWidget(mainLayout->count() - 1, dice);
	connect(dice, SIGNAL(deleteMe()), this, SLOT(removeDamageDice()));
}

void AttackDice::removeDamageDice()
{
	if (damageDiceList.count() > 1)
	{
	AttackDamageDice* dice = qobject_cast<AttackDamageDice*>(sender());
	int index = damageDiceList.indexOf(dice);
	delete damageDiceList.takeAt(index);
	}
	if (damageDiceList.count() <= 1)
		damageDiceList[0]->setDeleteEnabled(false);
	addDamageDiceButton->setEnabled(true);
}

StatsDice::StatsDice(QWidget *parent) : AbstractDice(parent)
{
	numberOfDice	= new KIntSpinBox(2, 999, 1, 4, this);
	diceSize		= new KIntSpinBox(1, 999, 1, 6, this);
	dLabel			= new QLabel(tr("d"));
	rejectLabel		= new QLabel(tr("Reject: "));
	rejectValue		= new KIntSpinBox(1, 3, 1, 1, this);
	rejectHL		= new QComboBox;

	rejectHL->addItem(tr("Lowest"));
	rejectHL->addItem(tr("Highest"));

	numberOfDice->setWhatsThis(tr("The number of dice to roll"));
	diceSize->setWhatsThis(tr("The number of sides the dice has<br><br>"
								   "d20 is equlivant to rolling a dice with 20 sides"
								   " producing a number between 1 and 20"));

	mainLayout = new QVBoxLayout;
	QHBoxLayout *hLayout1 = new QHBoxLayout;
	QHBoxLayout *hLayout2 = new QHBoxLayout;

	hLayout1->addWidget(numberOfDice, 1);
	hLayout1->addWidget(dLabel, 0);
	hLayout1->addWidget(diceSize, 1);
	hLayout2->addWidget(rejectLabel, 0);
	hLayout2->addWidget(rejectValue, 1);
	hLayout2->addWidget(rejectHL, 0);
	hLayout2->addWidget(rollButton, 0);
	hLayout1->addWidget(deleteButton, 0);
	mainLayout->addLayout(hLayout1);
	mainLayout->addLayout(hLayout2);
	setLayout(mainLayout);

	setFrameStyle(QFrame::Raised | QFrame::StyledPanel);

	connect(numberOfDice, SIGNAL(valueChanged(int)), this, SLOT(numberOfDiceChanged()));
}

void StatsDice::roll()
{
	int dice	= numberOfDice->value();
	int sides	= diceSize->value();
	QList<int> list(rollDice(dice, sides, 0));
	list.removeLast();
	list.removeFirst();

	int total = 0;
	int include = list.count() - rejectValue->value() - 1;
	if (rejectHL->currentText() == QString("Highest"))
		qSort(list);
	else
		qSort(list.begin(), list.end(), qGreater<int>());

	for (int i = 0; i <= include; ++i)
		total += list[i];

	QString text;
	text.append(tr("%1 = (").arg(total));

	for (int i = 0; i < include; ++i)
		text.append(tr("%1 + ").arg(list[i]));

	text.append(tr("%1) Rejected: ").arg(list[include]));

	for (int i = include + 1; i < list.count()-1; ++i)
		text.append(tr("%1, ").arg(list[i]));
	text.append(tr("%1").arg(list[list.count() - 1]));

	emit diceRolled(text);
}

void StatsDice::save(QDataStream& stream)
{
	stream << qint32(Stats);

	stream << qint32(numberOfDice->value());
	stream << qint32(diceSize->value());
	stream << qint32(rejectHL->currentIndex());
	stream << qint32(rejectValue->value());
}

void StatsDice::open(QDataStream& stream)
{
	qint32 value;
	stream >> value;
	numberOfDice->setValue(value);
	stream >> value;
	diceSize->setValue(value);
	stream >> value;
	rejectHL->setCurrentIndex(value);
	stream >> value;
	rejectValue->setValue(value);
}

void StatsDice::numberOfDiceChanged()
{
	rejectValue->setMaximum(numberOfDice->value() - 1);
}

WODDice::WODDice(QWidget *parent) : AbstractDice(parent)
{
	poolLabel = new QLabel(i18n("Dice Pool: "));
	pool = new KIntSpinBox(1, 999, 1, 1, this);
	againLabel = new QLabel(i18n("Again: "));
	again = new KIntSpinBox(8, 10, 1, 10, this);

	mainLayout = new QHBoxLayout();
	mainLayout->addWidget(poolLabel, 0);
	mainLayout->addWidget(pool, 1);
	mainLayout->addWidget(againLabel, 0);
	mainLayout->addWidget(again, 1);
	mainLayout->addWidget(rollButton, 0);
	mainLayout->addWidget(deleteButton, 0);

	setLayout(mainLayout);
}

void WODDice::roll()
{
	int sucesses = 0;
	int dice = pool->value();
	int reroll = 0;

	QVector< QList<int> > results;
	results.append(rollDice(dice, 10, 0));
	results[0].removeFirst();
	results[0].removeLast();

	while(true)
	{
		reroll = 0;
		foreach(int i, results[results.count() - 1])
		{
			if(i >= again->value())
				++reroll;
		}
		if(reroll <= 0)
			break;
		results.append(rollDice(reroll, 10, 0));
		results[results.count() - 1].removeFirst();
		results[results.count() - 1].removeLast();
	}

	foreach(QList<int> list, results)
	{
		foreach(int i, list)
		{
			if(i >= 8)
				++sucesses;
		}
	}

	QString text;
	text.append(i18n("Sucesses: "));
	text.append(tr("%1 =").arg(sucesses));

	foreach(QList<int> list, results)
	{
		text.append(tr(" ("));
		for (int r = 0; r < list.count() - 1; ++r)
		{
			text.append(boldOn(list[r], 8));
			text.append(", ");
		}
		text.append(boldOn(list[list.count() - 1], again->value()));
		text.append(tr(")"));
	}
	emit diceRolled(text);
}

void WODDice::save(QDataStream& stream)
{
	stream << qint32(WOD);

	stream << qint32(pool->value());
	stream << qint32(again->value());
}

void WODDice::open(QDataStream& stream)
{
	qint32 value;
	stream >> value;
	pool->setValue(value);
	stream >> value;
	again->setValue(value);
}

LabelDice::LabelDice(QWidget *parent) : AbstractDice(parent)
{
	nameLabel		= new KLineEdit(i18n("Type title here"));
	addButton		= new QPushButton(KIcon("list-add"), "", this);
	dicePickerDlg	= new DicePicker(this);
	diceList		= new DiceList(0);

	QHBoxLayout* hlayout = new QHBoxLayout();
	hlayout->addWidget(nameLabel, 1);
	hlayout->addWidget(addButton, 0);
	hlayout->addWidget(deleteButton, 0);

	
	mainLayout = new QVBoxLayout;
	mainLayout->setSpacing(0);
	mainLayout->addLayout(hlayout);
	mainLayout->addLayout(diceList);
	hlayout->setSpacing(5);

	connect(dicePickerDlg, SIGNAL(addDice(AbstractDice*)), this, SLOT(addDice(AbstractDice*)));
	connect(addButton, SIGNAL(clicked(bool)), dicePickerDlg, SLOT(show()));
	connect(diceList, SIGNAL(diceRolled(QString)), this, SIGNAL(diceRolled(QString)));
	connect(diceList, SIGNAL(diceRemoved()), this, SLOT(removeDice()));

//	dicePickerDlg->show();

	setLayout(mainLayout);
	setContentsMargins(0,5,0,0);
}

void LabelDice::save(QDataStream& stream)
{
	stream << qint32(DLabel);
	stream << nameLabel->text();
	diceList->save(stream);
}

void LabelDice::open(QDataStream& stream)
{
	QString text;
	stream >> text;
	nameLabel->setText(text);
	diceList->open(stream);

}

void LabelDice::addDice(AbstractDice* dice)
{
	setFrameStyle( QFrame::StyledPanel | QFrame::Raised);
//	dice->setMargin(0);
	diceList->addDice(dice);
}

void LabelDice::removeDice()
{
	if (diceList->count() <= 0)
		setFrameStyle(QFrame::NoFrame);
}
