/*
	Copyright 2009 Michael Daffin <james1479@googlemail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef DICE_H
#define DICE_H

#include <QFrame>
#include <QWidget>
#include <QPushButton>
#include <QCheckBox>
#include <QBoxLayout>

class QRegExpValidator;
class QString;
class QLabel;
class QVBoxLayout;
class QHBoxLayout;
class QComboBox;
class QDataStream;
template < class T > class QList;
class KIntSpinBox;
class KLineEdit;
class DicePicker;
class DiceList;

enum DiceType{Custem		= 1,
			  D20			= 2,
			  AttackDamage	= 3,
			  Attack		= 4,
			  Stats			= 5,
			  WOD			= 6,
			  DLabel		= 7};

/**
 * An abstract dice that all others are baised off
 */
class AbstractDice : public QFrame
{
	Q_OBJECT
public:
	explicit AbstractDice(QWidget *parent = 0);
	/**
	 * Enables and disables the delete button
	 */
	void setDeleteEnabled(bool b = true)
		{ deleteButton->setEnabled(b); }
	void setMargin(int margin) { mainLayout->setMargin(margin); }

public slots:
	/**
	 * Rolls the dice and emits diceRolled(QString) with the result
	 */
	virtual void roll() = 0;
	/**
	 * Saves the dice's current settings
	 * @param stream the output stream to save to
	 */
	virtual void save(QDataStream& stream) = 0;
	/**
	 * Loads the dice's saved settings
	 * @param stream the input stream to load from
	 */
	virtual void open(QDataStream& stream) = 0;

signals:

	/**
	 * Emited when a dice is rolled along with the formatted result
	 */
	void diceRolled(QString);
	/**
	 * Emited when the delete button is pressed to tell the parent of the dice to remove it from the list or window
	 */
	void deleteMe();

protected:
	QPushButton *rollButton;
	QPushButton *deleteButton;
	QBoxLayout* mainLayout;
};

/**
 * A generic #d# + # dice where # can be anynumber
 */
class CustemDice : public AbstractDice
{
	Q_OBJECT
public:
	/*
	 * @pram parent The parent of the dice
	 */
	CustemDice(QWidget *parent = 0);

public slots:
	void roll();
	void save(QDataStream& stream);
	void open(QDataStream& stream);

private:
	KIntSpinBox *numberOfDice;
	KIntSpinBox *numberOfSides;
	KIntSpinBox *modifier;
	QLabel *dLabel;
	QLabel *modLabel;

};

/**
 * A generic 1d20 + # where # can be anynumber
 * Provided for convience to the user, CustemDice can do the same and more.
 */
class D20Dice : public AbstractDice
{
	Q_OBJECT
public:
	/*
	 * @pram parent The parent of the dice
	 */
	D20Dice(QWidget *parent = 0);

public slots:
	void roll();
	void save(QDataStream& stream);
	void open(QDataStream& stream);

private:
	QLabel *label;
	KIntSpinBox *modifier;

};

/**
 * A sub dice used in the AttackDice.
 */
class AttackDamageDice : public AbstractDice
{
	Q_OBJECT
public:
	AttackDamageDice(QWidget *parent = 0);
	bool isCritable() { return critical->isChecked(); }

public slots:
	void roll() {}
	QList<int> rollDamage();
	void save(QDataStream& stream);
	void open(QDataStream& stream);

private:
	KIntSpinBox* numberOfDice;
	QLabel* dLabel;
	KIntSpinBox* diceSize;
	QLabel* modLabel;
	KIntSpinBox* modifier;
	QCheckBox* critical;
};

/**
 * Used to roll attacks in the style of dungeons and dragons
 * Can tell when you critical or roll a natural 1
 */
class AttackDice : public AbstractDice
{
	Q_OBJECT
public:
/*
 * @pram parent The parent of the dice
 */
	AttackDice(QWidget *parent = 0);

public slots:
	void roll();
	/**
	 * Adds a damage sub dice to the attack dice (up to a max of 5)
	 */
	void addDamageDice();
	/**
	 * Removes and deletes a damage sub dice
	 */
	void removeDamageDice();
	void save(QDataStream& stream);
	void open(QDataStream& stream);

private:
	QLabel* attackBonusLabel;
	KLineEdit* attackBonus;
	QLabel* damageLabel;
	QPushButton* addDamageDiceButton;
	QLabel* critLabel;

	QList<AttackDamageDice*> damageDiceList;

	QLabel* critRangeLabel;
	KIntSpinBox* critRange;
	KIntSpinBox* multiplier;

};

/**
 * A stats style dice designed for dungeons and dragons
 * It can roll #d# where # is anynumber and reject somany of the lowest or highest values
 * e.g. 4d6 reject 2 lowest will reject the two lowest rolled numbers and sum hte remaining two
 */
class StatsDice : public AbstractDice
{
	Q_OBJECT
public:
/*
 * @pram parent The parent of the dice
 */
	StatsDice(QWidget *parent = 0);

public slots:
	void roll();
	/**
	 * Changes the maximum number of rejected dice to one lower then the total number of dice
	 */
	void numberOfDiceChanged();
	void save(QDataStream& stream);
	void open(QDataStream& stream);

private:
	KIntSpinBox* numberOfDice;
	QLabel* dLabel;
	KIntSpinBox* diceSize;
	QLabel* rejectLabel;
	QComboBox* rejectHL;
	KIntSpinBox* rejectValue;
};

class WODDice : public AbstractDice
{
	Q_OBJECT
public:
	WODDice(QWidget *parent = 0);

public slots:
	void roll();
	void save(QDataStream& stream);
	void open(QDataStream& stream);

private:
	QLabel* poolLabel;
	KIntSpinBox* pool;
	QLabel* againLabel;
	KIntSpinBox* again;
};

class LabelDice : public AbstractDice
{
	Q_OBJECT
public:
	LabelDice(QWidget *parent = 0);

public slots:
	void roll() {}
	void save(QDataStream& stream);
	void open(QDataStream& stream);
	void addDice(AbstractDice* dice);
	void removeDice();

private:
	KLineEdit* nameLabel;
	DicePicker* dicePickerDlg;
	QPushButton *addButton;
	DiceList* diceList;

};

#endif // DICE_H
