/*
	Copyright 2009 Michael Daffin <james1479@googlemail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "mainwindow.h"
#include "dice.h"
#include "dicepicker.h"
#include "dicelist.h"

/* Qt */
#include <QWidget>
#include <QVBoxLayout>
#include <QDataStream>
#include <QTextStream>

/* kde */
#include <KApplication>
#include <KAction>
#include <KLocale>
#include <KActionCollection>
#include <KStandardAction>
#include <KFileDialog>
#include <KMessageBox>
#include <KIO/NetAccess>
#include <KSaveFile>
#include <KToolBar>

quint32 static MAGIC_NUMBER = 0x6e67646b;
qint32 static VERSION = 102;

MainWindow::MainWindow(QWidget *parent) : KXmlGuiWindow(parent),fileName(QString())
{
	QWidget *cWidget		= new QWidget;
	QVBoxLayout* mainLayout	= new QVBoxLayout;
	diceList				= new DiceList;
	results					= new KTextEdit;

	setObjectName("mainwindow");
	diceList->setObjectName("main-dice-list");

	mainLayout->setSpacing(0);

	results->setReadOnly(true);
	results->setWhatsThis(tr("This is where the results of the dice rolls are displayed"));

	mainLayout->addLayout(diceList);
	mainLayout->addWidget(results);

	cWidget->setLayout(mainLayout);
	setCentralWidget(cWidget);

	setupActions();
	setupGUI (static_cast<StandardWindowOption> (Default &
						(~StatusBar)));

//	toolBar("mainToolBar")->hide();
//	toolBar("dndToolBar")->hide();
//	toolBar("wodToolBar")->hide();

	connect(diceList, SIGNAL(diceRolled(QString)), this, SLOT(addResults(QString)));
}

void MainWindow::setupActions()
{
	KStandardAction::quit(this, SLOT(close()), actionCollection());
	KStandardAction::save(this, SLOT(saveFile()), actionCollection());
	KStandardAction::saveAs(this, SLOT(saveFileAs()), actionCollection());
	KStandardAction::open(this, SLOT(openFile()), actionCollection());

	KAction* clearAction = new KAction(this);
	clearAction->setText(i18n("&Clear"));
	clearAction->setIcon(KIcon("edit-clear"));
	actionCollection()->addAction("edit-clear", clearAction);

	KIcon addIcon("list-add");

	KAction* addDiceAction = new KAction(this);
	addDiceAction->setText(i18n("D&ice"));
	addDiceAction->setIcon(addIcon);
	actionCollection()->addAction("add-dice", addDiceAction);

	KAction* addDiceDefaultAction = new KAction(this);
	addDiceDefaultAction->setText(i18n("D&efault"));
	addDiceDefaultAction->setIcon(addIcon);
	actionCollection()->addAction("add-dice-default", addDiceDefaultAction);

	KAction* addDiceD20Action = new KAction(this);
	addDiceD20Action->setText(i18n("&D20"));
	addDiceD20Action->setIcon(addIcon);
	actionCollection()->addAction("add-dice-d20", addDiceD20Action);

	KAction* addDiceAttackAction = new KAction(this);
	addDiceAttackAction->setText(i18n("&Attack"));
	addDiceAttackAction->setIcon(addIcon);
	actionCollection()->addAction("add-dice-attack", addDiceAttackAction);

	KAction* addDiceStatsAction = new KAction(this);
	addDiceStatsAction->setText(i18n("&Stats"));
	addDiceStatsAction->setIcon(addIcon);
	actionCollection()->addAction("add-dice-stats", addDiceStatsAction);

	KAction* addDiceWODAction = new KAction(this);
	addDiceWODAction->setText(i18n("&World Of Darkness"));
	addDiceWODAction->setIcon(addIcon);
	actionCollection()->addAction("add-dice-wod", addDiceWODAction);

	KAction* addDiceLabelAction = new KAction(this);
	addDiceLabelAction->setText(i18n("&Add Label"));
	addDiceLabelAction->setIcon(addIcon);
	actionCollection()->addAction("add-dice-label", addDiceLabelAction);

	results->addAction(clearAction);
	results->setContextMenuPolicy(Qt::ActionsContextMenu);

	connect( addDiceDefaultAction, SIGNAL(triggered(bool)), diceList, SLOT(addDiceDefault()));
	connect( addDiceD20Action, SIGNAL(triggered(bool)), diceList, SLOT(addDiceD20()));
	connect( addDiceAttackAction, SIGNAL(triggered(bool)), diceList, SLOT(addDiceAttack()));
	connect( addDiceStatsAction, SIGNAL(triggered(bool)), diceList, SLOT(addDiceStats()));
	connect( addDiceWODAction, SIGNAL(triggered(bool)), diceList, SLOT(addDiceWOD()));
	connect( addDiceLabelAction, SIGNAL(triggered(bool)), diceList, SLOT(addDiceLabel()));
	connect( addDiceAction, SIGNAL(triggered(bool)), diceList, SLOT(showDiceDialog()));
	connect( clearAction, SIGNAL(triggered(bool)), results, SLOT(clear()));

}

void MainWindow::addResults(QString result)
{
	results->append(result);
}


void MainWindow::saveFileAs(const QString &outputFileName)
{
	KSaveFile file(outputFileName);
	if(!file.open(QIODevice::WriteOnly)) {
		KMessageBox::error(this, i18n("Failed to open %1: Unable to write file")
								 .arg(fileName));
		return;
	}
	QDataStream stream(&file);
	stream << MAGIC_NUMBER;
	stream << VERSION;

	diceList->save(stream);

	file.finalize();
	file.close();

	fileName = outputFileName;
}

void MainWindow::saveFileAs()
{
	 saveFileAs(KFileDialog::getSaveFileName(KUrl("~/default"), "*.dice|Dice Files"));
}

void MainWindow::saveFile()
{
	if(!fileName.isEmpty())
	{
		saveFileAs(fileName);
	}
	else
	{
		saveFileAs();
	}
}

void MainWindow::openFile()
{
	openFile(KFileDialog::getOpenFileName(KUrl("~/"), "*.dice|Dice Files"));
}

void MainWindow::openFile(const QString &inputFileName)
{
	if(inputFileName.isEmpty())
		return;

	QString tmpFile;
	if(KIO::NetAccess::download(inputFileName, tmpFile,
		 this))
	{
	QFile file(tmpFile);
	if (!file.open(QIODevice::ReadOnly))
	{
		KMessageBox::error(this, i18n("Failed to open %1: Unable to read file")
								 .arg(fileName));
		return;
	}

	QDataStream stream(&file);
	quint32 magic;
	stream >> magic;
	if (magic != MAGIC_NUMBER)
	{
		KMessageBox::error(this, i18n("Failed to open %1: Wrong file type")
								 .arg(fileName));
		return;
	}

	qint32 version;
	stream >> version;
	if (version > VERSION)
	{
		KMessageBox::error(this, i18n("Failed to open %1: File version is too new")
								 .arg(fileName));
		return;
	}

	results->clear();

	diceList->open(stream);

	fileName = inputFileName;

	KIO::NetAccess::removeTempFile(tmpFile);
	}
	else
	{
	KMessageBox::error(this,
		KIO::NetAccess::lastErrorString());
	}
}


